package lda_gibbs;

import java.util.*;
import java.io.Serializable;

/**
 * Records a counts per an index (such as word or topic) and a
 * total count.
 *
 * @author akyrola@cs.cmu.edu, 2010
 */
public class IndexedCounter implements Serializable {

    private static final long serialVersionUID = 1;
    

    private int total = 0;
    private int counts[] = null;


    protected IndexedCounter() {

    }

    public int getNumOfWords() {
        return counts.length;
    }

    public IndexedCounter(int estimatedSize) {
        counts = new int[estimatedSize];
    }

    public void increment(int id, int incr) {
        if (id >= counts.length) {
            synchronized (this)  {
                if (id >= counts.length) {
                    expand((int) (id * 1.2));

                }
            }

        }
        counts[id] += incr;
        total += incr;
    }

    private void expand(int maxid) {
        int newlength = (int) (maxid);
        System.out.println("Have to expand " + maxid);
        int[] newa = new int[newlength];
        System.arraycopy(counts, 0, newa, 0, counts.length);
        counts = newa;
    }


    public int getTotal() {
        return total;
    }

    public List<Integer> getTop(int n) {
        TreeSet<IdCountTuple> list = new TreeSet<IdCountTuple>(new Comparator<IdCountTuple>() {
            public int compare(IdCountTuple o1, IdCountTuple o2) {
                if (o1.count > o2.count) return -1;        // Descending order
                else if (o1.count < o2.count) return 1;
                return (o1.id > o2.id ? -1 : 1);
            }
        });
        int min = Integer.MIN_VALUE;
        for(int id=0; id<counts.length; id++) {
            int c = counts[id];
            if (c > 0) {
                if ((list.size() < n) || (c>min)) {
                    list.add(new IdCountTuple(id, c));
                    if (list.size() == n+1) {
                        list.pollLast();
                        min = list.last().count;
                    }
                }
            }
        }

        List<Integer> result = new ArrayList<Integer>(n);
        for(IdCountTuple ict:list) {
            result.add(ict.id);
        }
        return result;
    }

    /** Adds counts from another counter to this and zeroes
     *  that counter.
     * @param diffCounter
     */
    public synchronized void consolidate(IndexedCounter diffCounter) {
        long t = System.currentTimeMillis();

        for(int i=0; i<diffCounter.counts.length; i++) {
            int a = diffCounter.counts[i];
            if (a != 0) {
                this.increment(i, a);
            }
            diffCounter.counts[i] = 0;
        }
        diffCounter.total = 0;
    }


    // Since each processor is working on different words, we need not to
    // lock when updating individual counts, only for the totals!
    // NOTE: This makes both global and thread local counts unsync!!
    public void consolidateSyncOnlytotals(IndexedCounter diffCounter, boolean syncWords, int modulo, int moduloRing) {
        // Check capacity
        synchronized (this) {
            if (diffCounter.counts.length > this.counts.length) {
                expand(diffCounter.counts.length);
            }
        }

        if (syncWords) {
            // Update counts
            for(int i=moduloRing; i<diffCounter.counts.length; i+=modulo) {
                int a = diffCounter.counts[i];
                if (a != 0) {
                    this.counts[i] +=a;
                    diffCounter.counts[i] = 0;                    
                }
            }
        } else {
            synchronized (this) {
                this.total += diffCounter.total;
            }
            diffCounter.total = 0;
        }
    }


    /**
     * Computes symmetric log2 based KL divergence of P(w|t)
     * between two distributions. Note: smoothed density used.
     *  REf: Newmwan et al. 2006 (Journal of Intelligence Community Research and Development)
     * @param other
     * @return
     */
    public double symmetricKLDivergence(IndexedCounter other, int W, double beta) {
        double kl = 0;
        if (other.counts.length < this.counts.length) other.expand(this.counts.length);
        if (this.counts.length < other.counts.length) this.expand(other.counts.length);
        int L = this.counts.length;

        for(int i=0; i<L; i++) {
            double p1 = (this.counts[i] + beta) / (this.total + W * beta);
            double p2 = (other.counts[i] + beta) / (other.total + W * beta);

            kl += p1 * Math.log(p1/p2) + p2 * Math.log(p2/p1);
        }

        
        kl /= 2;
        kl /= Math.log(2);
        return kl;
    }


    static class IdCountTuple {
        int id;
        int count;

        IdCountTuple(int id, int count) {
            this.id = id;
            this.count = count;
        }

        public String toString() {
            return id + ":" + count;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            IdCountTuple that = (IdCountTuple) o;

            if (count != that.count) return false;
            if (id != that.id) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = id;
            result = 31 * result + count;
            return result;
        }
    }

    public int getCount(int id) {
        if (id  < counts.length) {
            return counts[id];
        } else {
            return 0;
        }

    }

}
