package lda_gibbs;

import java.io.*;
import java.util.Random;
import java.util.Arrays;
import java.util.Collections;
import java.util.ArrayList;

/**
 * Represents a document in LDA. Also contains sampling code.
 *
 * Code is "research code", i.e pretty ugly.
 *
 * @author akyrola@cs.cmu.edu, 2010
 */
public class LDADocument {

    private int[] words;
    private int[] topicsOfWords;
    private int[] topicCounter;
    private double[] wordProbs;
    private int totalTopics; // after burn-in, does not change

    public LDADocument(File documentFile)
    {
        try {
            DataInputStream dis = new DataInputStream(new BufferedInputStream(new FileInputStream(documentFile)));
            int size = dis.readInt();
            int totalNumOfWords = dis.readInt();

            ArrayList<Integer> _words = new ArrayList<Integer>(totalNumOfWords);
            topicsOfWords = new int[totalNumOfWords];
            int wordIdx = 0;
            for(int i=0; i<size; i++) {
                int wordId = dis.readInt();
                int wordCount = dis.readInt();
                for(int j=0 ; j<wordCount; j++) {
                    _words.add(wordId);
                    wordIdx ++;
                }
            }
            /* Shuffle the order of words to decrease parallel clashes */
            Collections.shuffle(_words);
            words = new int[_words.size()];
            wordProbs  = new double[_words.size()];

            int i = 0;
            for(int x:_words) words[i++] = x;
            dis.close();
        } catch (IOException ioe) {
            throw new RuntimeException(ioe);
        }
    }


    public void initializeCounts(int numOfTopics, IndexedCounter[] topicWordCounter) {
        Random rand = new Random(System.currentTimeMillis());

        topicCounter = new int[numOfTopics];

        int topicId = rand.nextInt(topicCounter.length);
        for(int wordIdx=0; wordIdx<words.length; wordIdx++) {
            topicsOfWords[wordIdx] = topicId;
            topicCounter[topicId]++;
            topicWordCounter[topicId].increment(words[wordIdx], 1);
            totalTopics++;

        }
    }

    /**
     *  Perform Collapsed Gibbs Sampling for each word with id % wordModulo == wordQ
     *  in this doc and assigns new topics to words.
     *  @param rand Random generator
     *  @param wordQ which modulo-ring to sample
     *  @param wordModulo modulo
     *  @param numOfTopics number of topics
     *  @param alpha hyperparameter alpha
     *  @param beta hyperparameter beta
     *  @param totalWords size of vocabulary
     */
    public void resampleTopics(Random rand,
                               int wordQ,
                               int wordModulo,
                               IndexedCounter[] topicWordCounter,
                               int numOfTopics,
                               double alpha,
                               double beta,
                               int totalWords){
        double[] samplingArray = new double[numOfTopics];

        // Initialize thread local counters
        getThreadCounterForTopic(0, numOfTopics, totalWords);
        IndexedCounter[] localCounters = threadLocalDiffCounter.get();



        for(int wordIndex = 0; wordIndex<words.length; wordIndex++) {
            int wordId = words[wordIndex];
            if ((wordId % wordModulo) == wordQ) {

                int currentTopic = topicsOfWords[wordIndex];
                double cumulative = 0;

                /* Compute P(z_i == j | z_(-1), w) for words which have v%wordModule = wordQ */
                for(int topicId = 0; topicId<numOfTopics; topicId++) {
                    double probTopic = probWordInTopic(topicWordCounter,
                            localCounters,
                            numOfTopics,
                            alpha, beta,
                            totalWords,
                            wordId,
                            topicId);
                    cumulative += probTopic;
                    samplingArray[topicId] = cumulative;

                    if (probTopic < 0) {
                        throw new RuntimeException("Probability was below zero!!");
                    }

                }

                /* Remove current assignment */
                if (currentTopic >= 0) {
                      localCounters[currentTopic].increment(wordId, -1);
                      topicCounter[currentTopic]--;
                      totalTopics--;
                }

                /* Sample */
                double x = rand.nextDouble() * cumulative;
                for(int topicId=0; topicId<numOfTopics; topicId++) {
                    if (x<=samplingArray[topicId]) {
                        wordProbs[wordIndex] = cumulative;
                        topicsOfWords[wordIndex] = topicId;

                        localCounters[topicId].increment(wordId, 1);
                        topicCounter[topicId]++;
                        totalTopics++;
                        break;
                    }
                }
            }
        }

        // Consolidate
        consolidateWithGlobalCounters(topicWordCounter, false, wordModulo, wordQ);
    }

    /**
     * Computes probability of topic given word.
     * @param topicWordCounter
     * @param localCounters
     * @param numOfTopics
     * @param alpha
     * @param beta
     * @param totalWords
     * @param wordId
     * @param topicId
     * @return
     */
    private double probWordInTopic(IndexedCounter[] topicWordCounter,
                                   IndexedCounter[] localCounters,
                                   int numOfTopics,
                                   double alpha,
                                   double beta,
                                   int totalWords,
                                   int wordId,
                                   int topicId) {
        int nWordInTopic = topicWordCounter[topicId].getCount(wordId) +
                localCounters[topicId].getCount(wordId);
        int totalWordsInTopic = topicWordCounter[topicId].getTotal() +
                localCounters[topicId].getTotal();
        int nTopicInDoc = topicCounter[topicId];
        double p =  (nWordInTopic + beta) / (totalWordsInTopic + totalWords * beta) *
                (nTopicInDoc + alpha) / (totalTopics + numOfTopics * alpha);

        return p;
    }


    public int[] getTopicsOfWords() {
        return topicsOfWords;
    }

    public int[] getWords() {
        return words;
    }

    /* P(w |d ) */
    public double[] getWordProbs() {
        return wordProbs;
    }

    public double getLogPerplexity() {
        double pplx = 0;
        for(double d : wordProbs) {
            pplx += Math.log(d);
        }
        return pplx;
    }



    // This code is very very ugly
    private IndexedCounter getThreadCounterForTopic(int topicId, int numOfTopics, int numOfWords) {
        if (threadLocalDiffCounter.get() == null) {
            IndexedCounter[] counters = new IndexedCounter[numOfTopics];
            for(int i=0; i<numOfTopics; i++) counters[i] = new IndexedCounter(numOfWords);
            threadLocalDiffCounter.set(counters);
        }
        return threadLocalDiffCounter.get()[topicId];
    }



    // Adds my counts to global counts and zeros my counts.
    public static void consolidateWithGlobalCounters(IndexedCounter[] globalCounters, boolean syncWords, int modulo, int moduloRing) {
        IndexedCounter[] counters = threadLocalDiffCounter.get();
        if (counters != null) {
            int start = new Random().nextInt(counters.length);
            for(int u=0; u<counters.length; u++) {
                int i = (u+start) % counters.length; // Try to avoid synchronizing on same topics with another thread
                globalCounters[i].consolidateSyncOnlytotals(counters[i], syncWords, modulo, moduloRing);
                assert(counters[i].getTotal() == 0);
            }
        }
    }




    static ThreadLocal<IndexedCounter[]> threadLocalDiffCounter =
            new ThreadLocal<IndexedCounter[]>() {
                @Override
                protected IndexedCounter[] initialValue() {
                    return  null;
                }
            };


}
