package lda_gibbs;


import java.util.List;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicInteger;
import java.io.*;

import lda_gibbs.datamgnt.WordIdMap;

/**
 * Main class for parallel gibbs implementation of LDA.
 * This is "research code", so pretty ugly.  Author is happy to provide help.
 *
 * Use WordSetPreprocessor to process a directory of text files to format the algorithm
 * understands.
 *
 * When run, following files are generated:
 * <ul>
 *    <li>Top words for each topic.</li>
 *    <li>Perplexity log.</ll>
 *    <li>Serialized file containing this class, which contains the posterior estimate.</li>
 * </ul>
 * @author akyrola@cs.cmu.edu, 2010
 */
public class LDAParameterEstimation implements Serializable {

    private static final long serialVersionUID = 1;


    private int numOfTopics;
    private IndexedCounter[] chainCounters;
    private IndexedCounter[] sampleCounters;
    private int burnInLength = 200;
    private int sampleLag = 100;
    private int numOfSamples = 8;
    private int perplexityInterval = 25;
    private double perplexity;

    int numOfCpus;

    double alpha, beta = 0.1;
    private WordIdMap wordIdMap;

    private transient ExecutorService executorService;
    private transient BufferedWriter perplexityLog;
    private static long discounts = 0;

    public LDAParameterEstimation(int numOfTopics,  WordIdMap wordIdMap, int numOfCpus) {
        this.numOfTopics = numOfTopics;
        this.alpha = 50.0 / numOfTopics;    // As per Griffiths, PNAT 2004
        this.wordIdMap = wordIdMap;
        this.numOfCpus = numOfCpus;
        executorService = Executors.newFixedThreadPool(numOfCpus);
    }


    public IndexedCounter[] getSampleCounters() {
        return sampleCounters;
    }


    public WordIdMap getWordIdMap() {
        return wordIdMap;
    }

    public void doSampling(File directory) throws Exception {
        long t = System.currentTimeMillis();
        initCounters();


        perplexityLog= new BufferedWriter(new FileWriter("perplexity_" +
                directory.getName() + "_" + System.getProperty("num_of_threads") + "_" + System.currentTimeMillis() + ".txt"));
        perplexityLog.write("t\tppl\n");

        List<LDADocument> documents = new LinkedList<LDADocument>();

        // Initialize documents
        long totalWordsInCorpus = 0;
        for(File f : directory.listFiles()) {
            if (f.getName().endsWith(".ldat") && !f.getName().startsWith(".")) {
                LDADocument d = new LDADocument(f);
                documents.add(d);
                totalWordsInCorpus += d.getWords().length;
            }
        }

        int totalWords = wordIdMap.getNumOfWords();
        System.out.println("Loaded " + documents.size() + " documents, total words = " + totalWords);

        System.out.println("::: Total words=" + totalWords);
        System.out.println("::: Beta = " + beta);
        System.out.println("::: Alpha = " + alpha);
        System.out.println("::: Topics = " + numOfTopics);
        System.out.println("::: W = " + wordIdMap.getNumOfWords());
        System.out.println("::: Total words =" + totalWordsInCorpus);
        System.out.println("::: Words per doc = " + (totalWordsInCorpus/documents.size()));
        System.out.println("::: N of docs = " + documents.size());

        /* INITIALIZE */
        /* Make "Graph" of documents. Note: there are no edges! */
        ArrayList<DocumentSet> docGraph = new ArrayList<DocumentSet>(numOfCpus);

        for(int c=0; c<numOfCpus; c++) {
            docGraph.add(c, new DocumentSet());
        }

        // Add documents to corresponding vertices.
        int idSeq = 0;
        for (LDADocument doc : documents) {
            doc.initializeCounts(numOfTopics, chainCounters);
            docGraph.get((idSeq++) % numOfCpus).addDocument(doc);
        }


        LDADocumentSampler sampler = new LDADocumentSampler(alpha, beta, numOfTopics, totalWords, chainCounters);



        /**** BURN IN *****/
        runLDA(burnInLength, docGraph, sampler);

        /*** MARKOV CHAIN ****/
        for(int sampleRounds=0; sampleRounds<numOfSamples; sampleRounds++) {
            runLDA(sampleLag, docGraph, sampler);

            // Take samples - this is done synchronously
            for (LDADocument doc : documents) {
                int[] words = doc.getWords();
                int[] wordTopics = doc.getTopicsOfWords();
                for(int i=0; i<words.length; i++) {
                    sampleCounters[wordTopics[i]].increment(words[i], 1);
                }
            }

            long st = System.currentTimeMillis();

            for(int topicId=0; topicId<numOfTopics; topicId++) {
                System.out.println("====== Topic: " +  topicId + " =======");
                List<Integer> topWordIds = sampleCounters[topicId].getTop(10);
                for(int wid:topWordIds) {
                    System.out.println(wordIdMap.getWord(wid));
                }
            }

            discounts += (System.currentTimeMillis() - st); // Dont measure

        }
        System.out.println("Sampling took " + (System.currentTimeMillis() - t) + " ms");
        executorService.shutdown();
    }

    public int totalIterations = 0;


    private transient ThreadLocal<Integer> samplingCpuModulo = new ThreadLocal<Integer>() {

        int j = 0;

        protected Integer initialValue() {
            synchronized (this) {
                this.set(j++);
            }
            return this.get();
        }


    };


    private void runLDA(int nIterations, final List<DocumentSet> docGraph, final LDADocumentSampler sampler) {
        if (nIterations % perplexityInterval != 0) throw new RuntimeException("Invalid iteration count");

        final int modulo = docGraph.size();

        for(int i=0; i<nIterations/perplexityInterval; i++) {

            for(int iter=0; iter<perplexityInterval; iter++) {
                long t = System.currentTimeMillis();


                // Instead of graphlab use executor

                final Object terminationLock = new Object();
                final AtomicInteger countDown = new AtomicInteger(numOfCpus);
                final boolean sync = (iter == perplexityInterval-1);
                synchronized (terminationLock) {
                    for (int j=0; j<numOfCpus; j++) {
                        executorService.submit(new Runnable() {
                            public void run() {
                                try {
                                    int vMod =  samplingCpuModulo.get();
                                    for(int q=0; q<numOfCpus; q++) {

                                        sampler.sample(  docGraph.get ((q+vMod)%modulo),
                                                vMod, modulo);

                                    }

                                    if (sync) {
                                        System.out.println("Synced: " + vMod + "/" + modulo);
                                        LDADocument.consolidateWithGlobalCounters(chainCounters, true, modulo, vMod);
                                    }
                                    if (countDown.decrementAndGet() == 0) {
                                        synchronized (terminationLock) {
                                            terminationLock.notifyAll();
                                        }
                                    }
                                } catch (Exception err){
                                    err.printStackTrace();
                                }
                            }
                        });
                    }

                    // barrier
                    try {
                        terminationLock.wait();
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }

                }
                System.out.println("one round took: " + (System.currentTimeMillis() - t) + " ms");

                totalIterations++;
                System.out.println("======== " + totalIterations + " ========");
            }
            computePerplexity(docGraph);
        }
    }

    public double getPerplexity() {
        return perplexity;
    }

    int iterCounter =0;

    private void computePerplexity(List<DocumentSet> docGraph) {

        // Do not include perplexity in time computation.
        long st = System.currentTimeMillis();

        int N = 0;
        double ppl = 0;
        for (DocumentSet ds : docGraph) {
            for (LDADocument document : ds.getDocuments()) {
                N += document.getWords().length;
                ppl += document.getLogPerplexity();
            }
        }
        this.perplexity = Math.exp(-1.0/(1.0*N)*ppl);
        System.out.println("In-sample perplexity: " + perplexity + " log ppl=" + ppl);
        iterCounter += perplexityInterval;
        try {
            perplexityLog.write(iterCounter + "\t" + perplexity + "\n");
            perplexityLog.flush();
        } catch (IOException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }

        discounts += (System.currentTimeMillis() - st);
    }

    private void writeTopWords(File outputFile) throws IOException {

        BufferedWriter writer = new BufferedWriter(new FileWriter(outputFile));

        for(int topicId=0; topicId<numOfTopics; topicId++) {
            writer.write("====== Topic: " +  topicId + " =======\n");
            List<Integer> topWordIds = sampleCounters[topicId].getTop(10);
            for(int wid:topWordIds) {
                double prob = sampleCounters[topicId].getCount(wid) * 1.0 / sampleCounters[topicId].getTotal();
                writer.write("\t" + wordIdMap.getWord(wid) + "\t" + prob + "\n");
            }
        }
        writer.close();

    }


    private void initCounters() {
        chainCounters = new IndexedCounter[numOfTopics];
        sampleCounters = new IndexedCounter[numOfTopics];
        for(int topicId=0; topicId<numOfTopics; topicId++) {
            chainCounters[topicId] = new IndexedCounter(1+wordIdMap.getNumOfWords());
            sampleCounters[topicId] = new IndexedCounter(1+wordIdMap.getNumOfWords());
        }
    }

    public static void main(String[] args) throws Exception {
        if (args.length < 3) {
            System.err.println("Usage: LDAParameterEstimation [directory of processed files][num-of-topics] [outputfile (top words)]");
        }

        File directory = new File(args[0]);
        int numOfTopics = Integer.parseInt(args[1]);
        File outputFile = new File(args[2] + "_" + System.getProperty("num_of_threads", "2")  );
        WordIdMap wordIds = new WordIdMap(new File(directory, "index.lidx"));

        long t = System.currentTimeMillis();

        int ncpus = Integer.parseInt(System.getProperty("num_of_threads", "2"));

        System.out.println("****** NUM OF THREADS: " + ncpus + " ******");

        LDAParameterEstimation estimator = new LDAParameterEstimation(numOfTopics, wordIds, ncpus);
        estimator.doSampling(directory);
        double timeInSecs = (System.currentTimeMillis() - t - discounts) * 0.001;

        estimator.writeTopWords(outputFile);

        System.out.println("Discounts in millisecs: " + discounts);


        // Output data
        String dirname = directory.getName();
        if (dirname.equals(".")) dirname = directory.getParentFile().getName();
        String fileName = dirname + "_" + ncpus + ".obj";
        FileOutputStream fos = new FileOutputStream(new File(fileName));
        ObjectOutputStream oos = new ObjectOutputStream(fos);
        oos.writeObject(estimator);
        oos.close();
        fos.close();
        fileName = dirname + "_" + ncpus + ".obj." + System.currentTimeMillis();
        fos = new FileOutputStream(new File(fileName));
        oos = new ObjectOutputStream(fos);
        oos.writeObject(estimator);
        oos.close();
        fos.close();



        // Write benchmark system output
        BufferedWriter wr = new BufferedWriter(new FileWriter(".runstats.R"));
        wr.write("execution_time=" + timeInSecs + "\nncpus=" + ncpus + "\n");
        wr.write("taskcount=1\nwork=0\nresidual=" + estimator.getPerplexity() + "\nmemory_writes_mb=0\nmemory_reads_mb=0\n");
        wr.close();
    }





    public int getBurnInLength() {
        return burnInLength;
    }

    public void setBurnInLength(int burnInLength) {
        this.burnInLength = burnInLength;
    }

    public int getSampleLag() {
        return sampleLag;
    }

    public void setSampleLag(int sampleLag) {
        this.sampleLag = sampleLag;
    }

    public int getNumOfSamples() {
        return numOfSamples;
    }

    public void setNumOfSamples(int numOfSamples) {
        this.numOfSamples = numOfSamples;
    }

    public double getAlpha() {
        return alpha;
    }

    public void setAlpha(double alpha) {
        this.alpha = alpha;
    }

    public double getBeta() {
        return beta;
    }

    public void setBeta(double beta) {
        this.beta = beta;
    }
}

