package lda_gibbs.datamgnt;

import java.util.*;
import java.io.*;
import java.text.NumberFormat;

/**
 * Takes a collection of documents, find all words used and gives them
 * a unique id. Converts documents into binary format. Some of the
 * non-informative words like "so", "if", "and" are pruned.
 *
 * Processed files are stored in directory "processed" inside the source dir.
 * @author akyrola@cs.cmu.edu, 2010
 */
public class WordSetPreprocessor {

    private Map<String, Integer> wordIds = new HashMap<String, Integer>(20000, 0.9f);
    private int idCounter = 1;

    private static HashSet<String> nonInformativeWords = new HashSet<String>();

    static {
        nonInformativeWords.addAll(Arrays.asList("a", "an", "the", "of", "in", "off", "out",
                "as", "then", "if", "so", "is", "not", "and", "or", "like", "we", "you", "i", "they",
                "us", "my", "your", "one", "yes", "but", "on", "it", "was", "still", "by", "to", "from", "with", "its", "too",
                "also", "do", "for", "are", "such", "when", "this", "that", "our", "ours", "their", "can", "may", "will",
                        "done", "them", "even", "over"));
    }

    public String normalize(String word) {
        return word.toLowerCase();
    }

    // TODO, accept words like SHA-1, JPEG2000
    public boolean isAcceptedWord(String s) {
        if (s.length() < 3) return false;
        if (nonInformativeWords.contains(s)) return false;
        for(char c : s.toCharArray()) {
            if (!Character.isLetter(c)) {
                return false;
            }
        }


        return true;
    }

    // TODO: filter 's, plurals, -ing, etc.
    private int getWordId(String word) {
        Integer wordId = wordIds.get(word);
        if (wordId == null) {
            wordId = (idCounter++);
            wordIds.put(word, wordId);
        }
        return wordId;
    }


    public Map<Integer, Integer> getWordCounts(String text) {
        text = text.toLowerCase();
        //TODO: line-delimiters (tavuviivat)

        HashMap<Integer, Integer> wordCounts = new HashMap<Integer, Integer>(3000, 0.9f);
        StringTokenizer zt = new StringTokenizer(text, " ,.:;-\"'`");
        int numberOfWords = 0;
        while(zt.hasMoreTokens()) {
            String word = zt.nextToken();
            if (isAcceptedWord(word)) {
                int wordId = getWordId(word);
                Integer count = wordCounts.get(wordId);
                if (count == null) count = 0;
                count++;
                wordCounts.put(wordId, count);
                numberOfWords++;
            }
        }
        wordCounts.put(-1, numberOfWords);
        return wordCounts;
    }


    public void writeWordcountFile(Map<Integer, Integer>wordCounts, File file) throws IOException {
        DataOutputStream dos = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(file), 65000));
        dos.writeInt(wordCounts.size()-1);
        dos.writeInt(wordCounts.get(-1)); // Total num of words encoded in key -1
        wordCounts.remove(-1);

        for(Map.Entry<Integer, Integer> tuple: wordCounts.entrySet()) {
            dos.writeInt(tuple.getKey());
            dos.writeInt(tuple.getValue());
        }
        dos.flush();
        dos.close();
    }

    public void writeWordIdFile(File file) throws IOException {
        DataOutputStream dos = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(file), 65000));
        dos.writeInt(wordIds.size());
        for(Map.Entry<String, Integer> tuple: wordIds.entrySet()) {
            dos.writeUTF(tuple.getKey());
            dos.writeInt(tuple.getValue());
        }
        dos.flush();
        dos.close();
    }


    public void processFile(File f, File outputFile) throws IOException {
        FileInputStream fis = new FileInputStream(f);
        byte[] b = new byte[(int) f.length()];
        int len = fis.read(b);

        if (len < b.length)
            throw new RuntimeException("Could not read file at once, fix the reader.");

        //System.out.println(f.getAbsolutePath());

        String str = preprocessString(new String(b));


        Map<Integer, Integer> wordCounts = getWordCounts(str);
        if (wordCounts.size() > 1)
            writeWordcountFile(wordCounts, outputFile);
    }

    private String preprocessString(String s) {
        int a = s.lastIndexOf("X-"); // email stuff
        if (a > 0) {
            int b = s.indexOf("\n", a);
            int c = s.indexOf("\r", a);
            b = Math.max(b,c);
            if (b>0)
                s = s.substring(a);
        }
        return s;

    }

    int numFiles = 0;
    long bytes = 0;

    public void processDirectory(File dir, File outputDir) throws IOException {
        for(File f : dir.listFiles()) {
            if (!f.isDirectory() && !f.getName().startsWith(".")) {
                this.processFile(f, new File(outputDir, numFiles + "_" + f.getName()  + ".ldat"));
                numFiles++;
                bytes += f.length();

                if (numFiles % 1000 == 0) System.out.println(numFiles + "/" + bytes + " bytes");
            } else if (f.isDirectory()) {
                processDirectory(f, outputDir);
            }
        }
    }


    public static void main(String[] args) throws Exception {
        if (args.length < 1) {
            System.err.println("Usage: WordSetPreProcessor [directory]");
        }

        long t = System.currentTimeMillis();

        File dir = new File(args[0]);
        File processedDir = new File(dir, "processed");
        WordSetPreprocessor processor = new WordSetPreprocessor();
        processedDir.mkdir();
        processor.processDirectory(dir, processedDir);
        processor.writeWordIdFile(new File(processedDir, "index.lidx"));

        t = System.currentTimeMillis() - t;
        NumberFormat nf = NumberFormat.getInstance();
        System.out.println("Processing " + processor.numFiles + " files took " + t + " ms.");
        System.out.println(nf.format(t*1.0/processor.numFiles) + " ms / file.");
        System.out.println(nf.format((processor.bytes*1.0/1024.0/1024.0) / (t*0.001)) + " MB/sec.");
        System.out.println(processor.idCounter + " unique words.");
    }


}
